/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_psm.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief Bluetooth protocol/service multiplexor
 * 
 * \section psm_mux Multiplexing
 * 
 * The protocol/service multiplexor is the entity used by
 * the connection-less transport layers (#l2cap_cl and #mhop_cl) to deliver
 * received packets to different application services.
 * 
 * Services on the receiving side are identified by a "PSM", a unique number
 * which is included in every data packet.
 * 
 * To receive data, application services have to register at the
 * protocol/service multiplexor, specifying the "PSM" to use. E.g.
\verbatim
// PSM your service uses
#define SERVICE_PSM 0x1009

// data reception funtion of your service
_service_data_rcv_cb(...) {
    ...
}

int main(void) {
	bt_psm_t* psmux;
	long service_nr;
	
	...
	
	// initialize protocol/service multiplexor
	psmux = bt_psm_init(bt_stack, 6, 6);

	// register your service
	servic_nr =
		bt_psm_service_register(psmux, SERVICE_PSM, _service_data_rcv_cb, NULL);
	
	...
}
\endverbatim
 * 
 * Data can now be sent to your service by sending it to the registered
 * psm using one of the connection-less transport protocols, e.g.
\verbatim
u_char data[12];
strcpy(data, "hello world");
l2cap_cl_send(data, <con_handle>, <psm>, 12);
\endverbatim
 * see #l2cap_cl or #mhop_cl for further details.
 * 
 * \section mux_buf Buffering
 * 
 * For buffering incoming packets, the protocol/service multiplexor
 * provides a set of packet buffers - the default message buffers. Buffered
 * packets are then delivered to the corresponding service in the context
 * of a special thread - the default message bearer thread - for that the
 * bt-stack is not blocked too long.
 * 
 * After registering a service at the protocol/service multiplexor, the
 * service is set up to use the default message buffers, as well as the default
 * message bearer.
 * 
 * To change this default policy, #bt_psm_service_set_buffers can be called.
 * 
 * \par Dedicated buffers
 * 
 * For special needs, e.g. because your service needs to queue packet buffers,
 * you can create dedicated buffers for that there are still buffers available
 * for other services. Use #bt_acl_pkt_queue_create to create dedicated
 * packet buffers.
 * 
 * \par Dedicated message bearer
 * 
 * For special need, e.g. because your service would block the default packet
 * bearer too long, a dedicated packet bearer thread can be created. This
 * is done by calling #bt_psm_msg_bearer_create.
 * 
 * An example is given below:
\verbatim
// Pointer to your buffers
bt_acl_pkt_buf* buf_que;

int main(void) {
	long retval;
	
	...
	
	// create your buffers
	buf_que = bt_acl_pkt_queue_create(bt_stack, 4);
	
	// create an own message bearer
	msg_bearer = bt_psm_msg_bearer_create(1024, "CMPktBuf");
	
	// set up your buffers
	retval = bt_psm_service_set_buffers(psmux, servic_nr, &buf_que, msg_bearer);
	if (retval) {
		printf("error buffer setup (%li)!\n", retval);
		return;
	}
}
\endverbatim
 * 
 * Note:
 * - #bt_psm_get_default_bufs returns a pointer to the default packet buffers
 * - #bt_psm_get_default_bearer returns a pointer to the default message bearer
 * Thus,
\verbatim
// own message bearer
msg_bearer = bt_psm_msg_bearer_create(1024, "ServBear");

// set up the buffers:
bt_psm_service_set_buffers(psmux, service_nr, bt_psm_get_default_bufs(psmux), msg_bearer);
\endverbatim
 * tells the protocol / service multiplexor to use the default packet buffers
 * for buffering an incoming packet, while the packet shall be delivered in
 * context of a dedicated message bearer thread.
 * 
 * \section unbuf_servs Unbuffered services
 * Call:
\verbatim
bt_psm_service_set_buffers(psmux, service_nr, NULL, NULL);
\endverbatim
to have an unbuffered service. Note that your service has to process incoming
packets rather quick for that an unbuffered scheme is an option: as long
as your service processes an incoming packet, the whole bt-stack is blocked!
 * 
 */
#ifndef _BT_PSM_H_
#define _BT_PSM_H_

#include <sys/mutex.h>
#include <bt/bt_acl_defs.h>
#include <bt/bt_msg_defs.h>

/**
 * \name Errors
 */
//@{
/** The psm specified is not valid */
#define BT_PSM_ERR_INVALID_PSM				-1

/** The psm specified is used by another registered service */
#define BT_PSM_ERR_PSM_IN_USE				-2

/** No more services can be registered to the connection-less data channel */
#define BT_PSM_ERR_SERVICE_PORTS_EXHAUSTED	-3

/** The service number passed does not identify a registered service */
#define BT_PSM_ERR_INVALID_SERVICE_NR		-5

/** The psm field in the data packet received is valid but there is no
 * service registered on that psm */
#define BT_PSM_ERR_NO_SERVICE_REGISTERED	-6

/** The service to deliver does not provide enough free buffers */
#define BT_PSM_ERR_BUF_OVERFLOW				-7
//@}

/**
 * \brief Data Callback Definition.
 *
 * This callback will be called whenever any user data is received on the
 * on a specific PSM. This is specified by the service registration routine.
 * 
 * \param msg Pointer to the received message
 * \param data Pointer to the payload of the message
 * \param len Size of the message's payload
 * \param service_nr Number of the service this callback corresponds to
 * \param arg Pointer to the (optional) callback argument.
 * 
 * \see #bt_psm_service_register
 */
#define BT_PSM_DATA_CB \
	bt_msg* (*data_cb) (bt_msg* msg, \
						u_char* data, \
						u_short len, \
						u_short service_nr, \
						void* cb_arg)

/*
 * \brief Definition of the message bearer type.
 */
typedef struct bt_psm_msg_bearer_s bt_psm_msg_bearer;

/**
 * \brief Definition of the protocol / service multiplexor type.
 */
typedef struct bt_psm_s bt_psm_t;

/**
 * \brief Returns a pointer to the default queue of free ACL packet buffers
 * 
 * \see #bt_psm_service_set_buffers
 */
inline bt_acl_pkt_buf** bt_psm_get_default_bufs(bt_psm_t* psmux);

/**
 * \brief Returns the number of the free ACL packet buffers available
 *
 * \see #bt_psm_service_set_buffers
 */
inline u_short bt_psm_get_nr_default_bufs(bt_psm_t* psmux);

/**
 * \brief Returns a pointer to the default message bearer
 * 
 * \see #bt_psm_msg_bearer
 */
inline bt_psm_msg_bearer* bt_psm_get_default_bearer(bt_psm_t* psmux);

/**
 * Initializes the protocol / service multiplexor. 
 * 
 * Note: The bt-stack, and the ACL layer, have to be initialized before
 * calling this function. Thus, make shure that:
 * - #bt_hci_init()
 * - #bt_acl_init()
 * have been called.
 * 
 * \param bt_stack Pointer to the bt-stack to use
 * \param nr_services Defines the maximal number of services the
 * protocol / service multiplexor can handle.
 * \param nr_bufs Number of default message buffers to use
 * 
 * \return Pointer to the created protocol / service multiplexor
 */
bt_psm_t* bt_psm_init(struct btstack* bt_stack,
						u_short nr_services,
						u_short nr_bufs);

/**
 * \brief Registers a higher layer protocol / service at the
 * protocol / service multiplexor.
 * 
 * After calling this function, messages received on the psm \p psm
 * are passed to the calling service. That is, each time a #bt_msg is
 * received, the registered service is signalled by calling the registered
 * data callback \p BT_PSM_DATA_CB.
 * 
 * \param psmux Pointer to the protocol / service multiplexor to use
 * \param psm The psm to register on
 * \param BT_PSM_DATA_CB Data callback function. Will be called each time
 * connection-less data arrives on the psm \p psm
 * \param cb_arg Callback argument
 */
long bt_psm_service_register(bt_psm_t* psmux, u_short psm, BT_PSM_DATA_CB, void* cb_arg);

/**
 * \brief Sets the buffering scheme of the service given.
 * 
 * See detailed description at the beginning of this documentation.
 *
 * \param mux Pointer to the protocol/service multiplexor
 * \param servic_nr Service number of the registered service
 * \param free_bufs Pointer to an initialized packet queue
 * (see #bt_acl_pkt_queue_create). Pass NULL if your service shall
 * use no buffers at all.
 * \param pkt_bearer Pointer to an initialized packet bearer
 * (see #bt_msg_bearer_create). Pass NULL if your service shall
 * use no message bearer.
 */
short bt_psm_service_set_buffers(bt_psm_t* mux,
								u_short service_nr,
								bt_acl_pkt_buf** free_bufs,
								bt_psm_msg_bearer* bearer);

/** \brief Clears a previousely registered service
 * 
 * Clears the registered service with service nr \p service_nr.
 * 
 * \param stack Pointer to the connection-less l2cap stack
 * \param service_nr Service number of the service to clear
 */
short bt_psm_service_clear(bt_psm_t* mux, u_short service_nr);

/** \brief Delivers a received message to a registered service
 * 
 * Delivers the received message \p msg to the service registered at
 * psm \p psm. The parameters \p data and \p len are passed
 * to the higher layer service via the registered data callback
 * (see #BT_PSM_DATA_CB). Thus, \p data should be a pointer to the payload
 * of the packet, while \p data_len should be set to the lenght of the
 * payload.
 * 
 * Note: applications typically do not call this function! This function is
 * by lower layer protocols (e.g. l2cap connection-less)
 * to deliver received messages to higher layer protocols / services
 * (e.g. multi-hop layer, application service).
 * 
 * \param mux Pointer to the protocol / service multiplexor
 * \param msg Pointer to the message to be delivered
 * \param data Pointer to the payload of the message
 * \param len Size of the payload
 * \param psm The psm the packet buffer will be delivered to
 * \param free_msg Reference to a free packet buffer pointer:
 * the protocol / service multiplexor will store a pointer
 * to a free message buffer in the reference passed.
 * 
 * \returns 0 or possible errors
 */
short bt_psm_deliver_msg(bt_psm_t* mux,
						bt_msg* msg,
						u_char* data,
						u_short len,
						u_short psm,
						bt_msg** free_msg);
							
/**
 * \brief Creates a message bearer.
 * 
 * By creating a message bearer for a given service, received messages are
 * queued. Delivery of the queued messages is then done by a newly created
 * "delivery" thread - a "message bearer".
 * 
 * Usually there's no need for creating a dedicated message buffer, as long
 * as the execution time of the registered callback function is very long.
 * 
 * \param stack Pointer to the connectionless l2cap stack
 * \param service_nr Number of the service a packet buffer shall be created for
 * \param buf_size Number of packets that can be stored in the packet buf
 * \param td_size Size of the created delivery thread
 * \param td_name A name for the created delivery thread
 */
bt_psm_msg_bearer* bt_psm_msg_bearer_create(u_short td_size, char* td_name);

/**
 * \brief Destroys a message bearer
 */
void bt_psm_msg_bearer_destroy(bt_psm_msg_bearer* bearer);

#endif //_BT_PSM_H_
