/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_msg_defs.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 */

#ifndef CM_XTC_DEFS_H_
#define CM_XTC_DEFS_H_

#include <bt/bt_defs.h>
#include <bt/bt_hci_api.h>
#include <utils/srtd_list.h>
#include <bt/bt_psm.h>
#include <cm/_con_mgr.h>

// set default debug level
#define LOG_LEVEL 0
#define LOG_CLASS 5
#include <debug/log_set.h>

#if LOG_LEVEL >= 3
#define CM_DEBUG_DUMP_NEIGHBOR_ORDER() _cm_dbg_dump_neighbor_order()
#define CM_DEBUG_DUMP_INQRES() _cm_dbg_dump_inqres()
#define CM_DEBUG_DUMP_CON_BUF() _cm_dbg_dump_con_buf()
#define CM_DEBUG_REG_CMDS() cm_xtc_register_dbg_cmds()
#else
#define CM_DEBUG_DUMP_NEIGHBOR_ORDER()
#define CM_DEBUG_DUMP_INQRES()
#define CM_DEBUG_DUMP_CON_BUF()
#define CM_DEBUG_REG_CMDS()
#endif



typedef struct _device_params {
    u_short clock_offset;
    u_char page_scan_rep_mode;
} cm_dev_params_t;

/// Neighbor Parameters
/** This struct defines a neighbor device and the parameters of the corresponding
 * connection.*/
typedef struct _cm_neighbor {
    bt_addr_t addr;
    signed char rssi;
    cm_dev_params_t dev_params;
    bt_hci_con_handle_t con_handle; ///< Connection Handle
    /**< If there is a connection to this neighbor device, con_handle is the
     * corresponding connection handle. If there's no connection to this
     * neighbor device, con_handle is set to BT_HCI_HANDLE_INVALID.*/
    struct _cm_neighbor* better_com_neigh; ///< Better Common Neighbor
    struct _cm_neighbor* next;      ///< used for con_buf linked list
    /**
     * nr. of inquiries since neighbor was last seen, e.g. 0 if it was seen
     * at last inquiry. Used for expiration of neighbors.
     */
    u_char last_seen; ///< when was this neighbor last seen 
    /**
     * if not null, xtc will not connect to this neighbor. Pointer to a neighbor
     * that is nearer.
     */
    struct _cm_neighbor* reduced; ///< reduction pointer
} cm_neigh_t;

/// Connection manager stack
typedef struct _cm_stack {
    bt_addr_t my_addr;              ///< BT address of this device
    struct bt_hci_inquiry_result* inqres;
                                    ///< Pointer to the inquiry results
    u_char nr_inq_devs;             ///< Number of devices found during latest inquiry
    u_long nr_inqs;                 ///< Number of inquiries done so far
    srtd_list_t* order;
    cm_neigh_t** neighbors;         ///< Neighbor order (sorted list containing
                                    ///< discoverd neighbor devices)
    u_char nr_neighbors;            ///< Number of neighbor devices in neighbor order
    cm_neigh_t con_buf;
    u_char data_buf[115];
                                    ///< Data buffer to send packets
    MUTEX data_mutex;               ///< Data mutex
    MUTEX bt_module;                ///< BT-module mutex
//    HANDLE start_new_inq;           ///< Handle to (re)start the inquiry thread
    HANDLE con_mgr_start;           ///< Handle to start connection management
                                    ///< thread
    struct btstack *bt_stack;       ///< Pointer to lower level hci stack 
    bt_psm_t* psmux;
    u_short psm;
    bt_acl_pkt_buf* buf_que;		///< Pointer to the packet buffers
    u_long cod;
                                    ///< Pointer to the rel con change cb
    u_long xtc;                   ///< Turns on|off XTC link selection
    u_long ai_max_period;         ///< Inquiry period after startup phase [min]
    u_long ai_min_period;         ///< Startup inquiry period [sec]
    u_long ai_init;               ///< Number of inquiries during initialization phase
    u_long ai_trans;              ///< Number of inquiries during transition phase
    u_long ai_period_deviation;   ///< Maximal deviation of inquiry period
    u_long ai_time;               ///< Duration [sec] of an inquiry
    u_long autinq;                ///< To turn on/off automatic neighbor discovery
    u_long rssi_id;               ///< Minimal RSSI difference (Id mode)
    u_long rssi_th;               ///< RSSI updating threshold
    u_long con_period;            ///< Maximal wait time between two connection
                                        ///< establishment trials
    u_long min_vtime;             ///< Minimal visibility time between inq/con establ
    u_long lpf_alpha;             ///< LPF coefficient alpha
    u_long lpf_beta;              ///< LPF coefficient beta
    u_long ptime;                 ///< Page timeout
} cm_stack_t;

/// Pointer to the CM Stack
cm_stack_t* cm_stack;

const char* _cm_dbg_rel_cons_to_string(cm_neigh_t** order, u_char size);
void _cm_dbg_dump_inqres(void);
void _cm_dbg_dump_con_buf(void);
void _cm_dbg_dump_neighbor_order(void);
void cm_xtc_register_dbg_cmds(void);


extern char _cm_dbg_buf[];

#endif /*CM_XTC_DEFS_H_*/
