#include <sync/sync.h>
//#include <sync/sync-anchor.h>
#include <stdio.h>
//#include <stdlib.h>				//only for debug printing (we need abs() in _print_dec)
#include <sys/atom.h>
#include <sys/timer.h>
#include <led/btn-led.h>

 /*******************************************************************************
 * _lltobs
 ******************************************************************************/
/**
 * @brief Internal function which converts a long long to a byte string
 * This function is used to convert a bound into characters for transmission.
 * 
 */
// ATTENTION: n will become unusable, if SYNC_BOUND_LEN != 6
int _lltobs(u_longlong * n, u_char * ps) {

	u_char * ptr = (u_char *) n;
	ps[0] = *(ptr + 0);
	ps[1] = *(ptr + 1);
	ps[2] = *(ptr + 2);
	ps[3] = *(ptr + 3);
	ps[4] = *(ptr + 4);
	ps[5] = *(ptr + 5);

	return 0;
}

 /*******************************************************************************
 * _bstoll
 ******************************************************************************/
/**
 * @brief Internal function which converts a byte string to a long long
 * 
 * This function is used to convert characters from transmission into a bound.
 */
void _bstoll(u_char * ps, u_longlong * n){

	ps[SYNC_BOUND_LEN - 1] &= (u_char) 0x03;

	u_char * ptr = (u_char *) n;
	*(ptr + 0) = ps[0];
	*(ptr + 1) = ps[1];
	*(ptr + 2) = ps[2];
	*(ptr + 3) = ps[3];
	*(ptr + 4) = ps[4];
	*(ptr + 5) = ps[5];
	*(ptr + 6) = (u_char) 0;
	*(ptr + 7) = (u_char) 0;

}

/*******************************************************************************
 * _print_dec
 ******************************************************************************/
/**
 * @brief Prints out a long long as decimal number
 * 
 * This function can be used for debugging as printf() doesn't support
 * the long long integer type.
 */
void _print_dec(long long n) {
	/* print in decimal */
	char chars[32];
	int count = 0;
	if(n < 0){
		printf_P(PSTR("-"));
	}
	do{
		chars[count] = abs(n % 10);
		n = n / 10;
		count = count + 1;
	}while(n);
	count = count - 1; /* just incremented above */

	while(count >= 0){
		printf_P(PSTR("%d"), chars[count]);
		count = count - 1;
	}
}

 /*******************************************************************************
 * sync_assemble_timestamp
 ******************************************************************************/
/**
 * @brief Converts two bounds into a timestamp stored in sync_timestamp_to_send
 */
void _sync_assemble_timestamp(syncbounds * bounds, u_char* stamp)
{
	_lltobs(&(bounds->lower), stamp);
	_lltobs(&(bounds->upper), stamp + SYNC_TSTAMP_OFFSET_1);	
	stamp[SYNC_TSTAMP_OFFSET_1 - 1] |= (u_char) (bounds->validity << 4);
}

 /*******************************************************************************
 * sync_dissassemble_timestamp
 ******************************************************************************/
/**
 * @brief Extracts the two bounds from the sync_timestamp_rcvd timestamp
 */
void _sync_dissassemble_timestamp(u_char* stamp, syncbounds* bounds)
{
	if((bounds->validity = (stamp[SYNC_TSTAMP_OFFSET_1 - 1]) >> 4) == 1){
		_bstoll(stamp, &(bounds->lower));
		_bstoll(stamp + SYNC_TSTAMP_OFFSET_1, &(bounds->upper));		
	}
}

/*******************************************************************************
 * _sync_calculate_drift
 ******************************************************************************/
/**
 * @brief Internal function that performs a division by 2^20.
 * 
 * This function is called by get_updated_bounds an computes current bounds
 * from the bounds stored according to the formula given by the drift-constraint
 * model.
 * 
 * As the ATMega's instruction set doesn't offer a division, this done by an
 * arithmetic right-shift. In order to get correctly rounded results, we first
 * add 2^20 to the operand, because this bit is going to be the MSB of the bits
 * neglected after the shift operation.
 */
void _sync_calculate_drift(long long * delta_h, u_longlong * bound, long int drift_fac){
	*bound += (long long) ((long long) (*delta_h * drift_fac) + 1048576) >> 20;
}

/*****************************************
 * get_updated_bounds                      *
 *****************************************/

syncbounds get_updated_bounds(syncbounds * bounds, u_long reference_time){
	//bounds to be returned; initialized with the current bounds
	syncbounds retval_bounds = *bounds;

	if (sync_globals.is_anchor) {
//		u_char stamp[SYNC_TSTAMP_LEN];
//		//anchor nodes don't do drift calculations, but get new bounds from pc
//		sync_anchor_get_timestamp(stamp);		
//		_sync_dissassemble_timestamp(stamp, &retval_bounds);
//		//we synchronize the led here, because this is the only case an
//		//	anchornode has valid bounds in dissassemled form.
//		btn_led_synch_offset = RT_COARSE_AVERAGE(retval_bounds.lower, retval_bounds.upper) - RT_NOW;
	}
	else {	
		long long delta_h;
		//time elapsed since the last update of the bounds
		delta_h = reference_time - bounds->last_update;
	
		//calculate new bounds
		if(delta_h < 0){
			_sync_calculate_drift(&delta_h, &retval_bounds.lower, MAX_DRIFT_FAC_BW);
			_sync_calculate_drift(&delta_h, &retval_bounds.upper, MIN_DRIFT_FAC_BW);
		}else{
			_sync_calculate_drift(&delta_h, &retval_bounds.lower, MIN_DRIFT_FAC);
			_sync_calculate_drift(&delta_h, &retval_bounds.upper, MAX_DRIFT_FAC);
		}
	}
	return retval_bounds;
}

/*******************************************************************************
 * sync_sync_bounds
 ******************************************************************************/
/**
 * @brief Intersect received bounds with local bounds
 *
 * The local bounds are updated to the current time and then intersected with the
 * received bounds. If the local bounds are not valid, rt_set_clock is called instead.
 * Invalid received bounds are ignored.
 *
 * Bounds are intersected by taking the maximum of the two lower bounds and the minimum
 * of the two upper bounds.
 *
 * On an anchornode nothing is done as received bounds are ignored.
 *
 * @param bounds_rcvd The bounds received
 * @param arrival_time The local time, when the new bounds arrived (or at least
 * an estimate of that time)
 */
void _sync_sync_bounds(syncbounds * bounds_rcvd, u_long arrival_time){

	//disjoint intervals: received bounds lay beneath our bounds
	//	=> discard bounds
	if(sync_globals.global_bounds.lower >= bounds_rcvd->upper){
		return;
	}
	//get current bounds, store them and
	sync_globals.global_bounds = get_updated_bounds(&sync_globals.global_bounds, arrival_time);
	//	set last_update to the right value.
	sync_globals.global_bounds.last_update = arrival_time;
	bounds_rcvd->upper += SYNC_DELAY_UNCERTAINTY;
	//disjoint intervals: received bounds lay above our bounds
	//	=> make our interval the union of the two intervals
	if(sync_globals.global_bounds.upper <= bounds_rcvd->lower){
		sync_globals.global_bounds.upper = bounds_rcvd->upper;
	}else{
		//lower_bound = max(lower_bound, lower_bound_received)
		if(sync_globals.global_bounds.lower < bounds_rcvd->lower){
			sync_globals.global_bounds.lower = bounds_rcvd->lower;
		}
		//upper_bound = min(upper_bound, lupper_bound_received)
		if(sync_globals.global_bounds.upper > bounds_rcvd->upper){
			sync_globals.global_bounds.upper = bounds_rcvd->upper;
		}
	}
	//synchronized heartbeat
	btn_led_synch_offset = RT_COARSE_AVERAGE(sync_globals.global_bounds.lower, sync_globals.global_bounds.upper) - RT_NOW;
}

/*************************************************
 * high-level functions (application interface)  *
 *************************************************/

char rt_is_valid(void){
	return SYNC_CURRENT_BOUNDS.validity;
}

syncbounds rt_get_bounds(void){
	return get_updated_bounds(&SYNC_CURRENT_BOUNDS, RT_NOW + SYNC_COMPUTATIONAL_DELAY);
}

void rt_get_average(u_longlong * av){
	syncbounds bounds = get_updated_bounds(&SYNC_CURRENT_BOUNDS, RT_NOW + SYNC_COMPUTATIONAL_DELAY);
	if(bounds.validity == 0){
		//rt clock hasn't been set yet: return the lower bound (this is a valid
		//	lower guaranteed rt-bound and is monotonicaly increasing.
		*av = bounds.lower;			
	}else{
		//compute the average with a leftshift
		*av = (bounds.lower + bounds.upper) >> 1;
	}
}

void rt_get_monotonic(u_longlong * millis){
	syncbounds bounds = get_updated_bounds(&SYNC_CURRENT_BOUNDS, RT_NOW + SYNC_COMPUTATIONAL_DELAY);
	//the lower guaranteed rt-bound is monotonicaly increasing.
	*millis = bounds.lower;			
}

int gettimeofday(timeval * tv, void * tz){
	//as tz is obsolete, we neglect it. just pass NULL for the second argument.
	u_longlong millistamp;
	rt_get_average(&millistamp);
	tv->tv_sec = (long) (millistamp / 1000);
	//granularity is only milliseconds
	tv->tv_usec = (long) ((millistamp % 1000) * 1000);
	return rt_is_valid() - 1;
}

u_long rt_get_millis(void){
	u_longlong ms;
	rt_get_average(&ms);
	return (u_long) ms;
}

u_long rt_get_seconds(void){
	u_longlong sec;
	rt_get_average(&sec);
	return (u_long) (sec / 1000);
}

/*****************************************
 * rt_init, rt_set_clock                 *
 *****************************************/

//should be done at startup
//	(especially sync_globals.global_bounds.validity has to be zero in order to
//	avoid unnecessary (and false) computations. furthermore, we want...
void rt_init(void){
	if (sync_globals.is_anchor) {
		sync_globals.global_bounds.validity = 1;		//the anchor node's bounds are always valid
		sync_globals.bitpattern = 0xA;
		btn_led_heartbeat(31, (char) sync_globals.bitpattern, 1);
	}
	else {
		sync_globals.bitpattern = 0x5;
		btn_led_heartbeat(10, (char) 0x1, 1);
	}
}

void rt_set_clock(syncbounds * current_bounds_rcvd)
{
	// only set clock of sensor nodes
	if (!sync_globals.is_anchor) {
		sync_globals.global_bounds.validity = 1;
		sync_globals.global_bounds.lower = current_bounds_rcvd->lower;
		sync_globals.global_bounds.upper = current_bounds_rcvd->upper + SYNC_DELAY_UNCERTAINTY;
		sync_globals.global_bounds.last_update = RT_NOW - SYNC_SET_DELAY;
							
		//change heartbeat after setting rt clock
		btn_led_synch_offset = RT_COARSE_AVERAGE(SYNC_CURRENT_BOUNDS.lower, SYNC_CURRENT_BOUNDS.upper) - RT_NOW;
		btn_led_heartbeat(31, (char) sync_globals.bitpattern, 1);
	}
}

/*****************************************
 * sync_generate_timestamp                 *
 *****************************************/
// handles request from 'bt_hci_transport_uart'
void sync_generate_timestamp(u_char* stamp)
{
	syncbounds send_bounds;
	
	if (sync_globals.is_anchor) {
		u_long t_curr;
		// generate timestamp out of current local time
		t_curr = NutGetMillis();
		send_bounds.lower = (u_longlong)t_curr;
		send_bounds.upper = (u_longlong)t_curr;
		send_bounds.last_update = t_curr;
		send_bounds.validity = 1;
	}
	else {
		// call to algorithm functions
		send_bounds = get_updated_bounds(&sync_globals.global_bounds, RT_NOW + SYNC_COMPUTATIONAL_DELAY);
	}
	_sync_assemble_timestamp(&send_bounds, stamp);
}

/*****************************************
 * sync_process_timestamp                *
 *****************************************/
// handles request from 'bt_hci_transport_uart'
void sync_process_timestamp(u_char* stamp, u_long t_arrive)
{
	syncbounds bounds;
	#ifdef MEASURE_MSG_DLY
		printf("#");	
	#endif
	_sync_dissassemble_timestamp(stamp, &bounds);
	if (!sync_globals.is_anchor) {
		//discard invalid synchronization data
		if(bounds.validity == 1){
			if(sync_globals.global_bounds.validity == 0){
				// we've got no valid bounds so far, so we initialize the
				// rtc using the bounds we've just received
				rt_set_clock(&bounds);
			}else{
				_sync_sync_bounds(&bounds, t_arrive - SYNC_SYNC_DELAY);
			}
		}
	}
}
