/*
 * Copyright (C) 2000-2005 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file wsnlab/wsnlab.c
 *
 * \date 16.05.2006
 *
 * \author bo
 *
 * \brief Quick hack library for the WSNLab2006
 *
 */
 
#include <stdio.h>
#include <string.h>
#include <sys/mutex.h>
#include <sys/timer.h>
#include <led/btn-led.h>
#include <cc/ccc.h>
#include <sys/atom.h>
#include <dev/irqreg.h>
#include <dev/twif.h>
#include <hardware/btn-hardware.h>
#include <wsnlab/wsnlab.h> 
#include <btsense/btsense.h> 

// ----------------------------------------------------------------------------
// #define constants
// ----------------------------------------------------------------------------

#define DICT_LENGTH 10
#define BYTES_PER_LINE 16
#define BTSENSE_I2C_TC74 0x48

// ----------------------------------------------------------------------------
// local variables
// ----------------------------------------------------------------------------

static MUTEX dict_mutex;
static pkt_element_t packet_dict[DICT_LENGTH];

// ----------------------------------------------------------------------------
// helper functions
// ----------------------------------------------------------------------------

/**
 * Prints a string representation of the data to stdout (like a hex editor)
 */
void print_data( char* data, u_short length )
{
	u_short idx = 0, jdx = 0;
	u_short numLines = (length / BYTES_PER_LINE) + (length % BYTES_PER_LINE > 0 ? 1 : 0);
	
	for ( idx = 0; idx < numLines; idx++ )
	{		
		printf( "%04u   ", idx * BYTES_PER_LINE );
	
		for ( jdx = idx*BYTES_PER_LINE; jdx < (idx+1)*BYTES_PER_LINE; jdx++ )
		{
			if ( jdx-idx*BYTES_PER_LINE == BYTES_PER_LINE / 2 )
			{
				printf( " " );
			}				
			
			if ( jdx < length )
			{
				printf( "%02X ", data[jdx] );
			}
			else
			{
				printf( "   " );
			}			
		}
		
		printf( "  " );
		for ( jdx = idx*BYTES_PER_LINE; jdx < (idx+1)*BYTES_PER_LINE; jdx++ )
		{
			if ( jdx < length )
			{			
				printf( "%c", data[jdx] );
			}
		}
		printf( "\n" );	
	}
	
	printf( "Length of data block: %u bytes\n", length );
}

void print_packet_short( ccc_packet_t* pkt )
{
	if ( pkt != NULL )
	{
		printf( "%04X | %04X |   %02X |   %4u | %s\n",
		pkt->dst, pkt->src, pkt->type, pkt->length, pkt->data );
	}
	else
	{
		printf( " *** NO PACKET ***\n" );
	}
}

inline u_short pow2(u_short e)
{
    return 1 << e;
} 

// FIXME: This could be done with a single hw write operation
void led_set_binary(u_short i)
{
	u_short e=3;
	btn_led_clear(0);
	btn_led_clear(1);
	btn_led_clear(2);
	btn_led_clear(3);
	while (i)
	{
		while(i < pow2(e)) e--;
		btn_led_set(e);
		i -= pow2(e);
	}
}

// TODO: FIXME!
int ccc_send_packet( u_short dest, u_char type, u_char* data, u_short length )
{
	int res = 0;
	ccc_packet_t *pkt = new_ccc_packet( length );
	
	pkt->length = length;
	memcpy( pkt->data, data, length );
	
	res = ccc_send( dest, type, pkt );
	
	free( pkt );
	
	return res;
}

void beep()
{
	btsense_sound(440);
    NutSleep(1000);
    btsense_sound(0);
}

// ----------------------------------------------------------------------------
// Dictionary
// ----------------------------------------------------------------------------

/* Note: A dictionary has 3 operations: 
 * 1. is_member(k)
 * 2. insert(k)
 * 3. delete(k)
 */

void dict_init(void)
{
	int idx;
	
	NutMutexLock( &dict_mutex );		
	for ( idx = 0; idx < DICT_LENGTH; idx++ )
	{
		packet_dict[idx].send_counter = 0;
		packet_dict[idx].recv_counter = 0;		
		packet_dict[idx].received_at  = 0;
		packet_dict[idx].pkt = NULL;
	}
	NutMutexUnlock( &dict_mutex );			
}

void dict_print(void)
{
	int idx;
	
	NutMutexLock( &dict_mutex );		
	for ( idx = 0; idx < DICT_LENGTH; idx++ )
	{		
		printf( "#%02u | send: %u | recvd: %u | at: %lu | ago: %lu \n", 
		        idx, 
		        packet_dict[idx].send_counter, 
		        packet_dict[idx].recv_counter,
		        packet_dict[idx].received_at,
		        NutGetMillis() - packet_dict[idx].received_at );		        
		print_packet_short( packet_dict[idx].pkt );
	}
	NutMutexUnlock( &dict_mutex );			
}

int dict_packet_is_known( ccc_packet_t *pkt )
{
	NutMutexLock( &dict_mutex );
	int idx;
	
	for ( idx = 0; idx < DICT_LENGTH; idx++ )
	{
		if ( packet_dict[idx].pkt != NULL && 
		     memcmp( pkt->data, packet_dict[idx].pkt->data, pkt->length ) == 0 )
		{
			// update the receive counter
			packet_dict[idx].recv_counter++;
	
			printf( "Received duplicate packet at %u: ", idx );
			print_packet_short( pkt );
        
			NutMutexUnlock( &dict_mutex );
			//return idx;
			return 1;
		}
	}
	
	NutMutexUnlock( &dict_mutex );
	return 0;
}

int dict_add_packet( ccc_packet_t *pkt )
{
	NutMutexLock( &dict_mutex );
	
	int idx = 0;
	
	while ( idx < DICT_LENGTH && packet_dict[idx].pkt != NULL )
	{
		idx++;
	}
	
	if ( idx < DICT_LENGTH && packet_dict[idx].pkt == NULL )
	{
		// we found a place in the send queue, so we add the current packet
		packet_dict[idx].received_at  = NutGetMillis();
		packet_dict[idx].recv_counter = 0;
		packet_dict[idx].send_counter = 0;
		
		// copy the packet in a local buffer
		packet_dict[idx].pkt = new_ccc_packet( pkt->length );
		memcpy( packet_dict[idx].pkt, pkt, sizeof(ccc_packet_t) + pkt->length );		
	}
	
	NutMutexUnlock( &dict_mutex );	
	
	return idx;
}

void dict_remove_packet( int idx )
{
	packet_dict[idx].send_counter = 0;
	packet_dict[idx].recv_counter = 0;	
	packet_dict[idx].received_at = 0;
	free( packet_dict[idx].pkt );
	packet_dict[idx].pkt = NULL;
}

pkt_element_t* dict_get_first_packet( int* handle )
{
	int idx = 0;
	NutMutexLock( &dict_mutex );
	
	while ( idx < DICT_LENGTH && packet_dict[idx].pkt == NULL )
	{
		idx++;
	}
	
	if ( idx < DICT_LENGTH )
	{
		*handle = idx;
		NutMutexUnlock( &dict_mutex );			
		return &packet_dict[idx];
	}
	
	NutMutexUnlock( &dict_mutex );	
	
	return NULL;
}

pkt_element_t* dict_get_next_packet( int* handle )
{
	int idx = *handle + 1;	
	NutMutexLock( &dict_mutex );	
	
	while ( idx < DICT_LENGTH  && packet_dict[idx].pkt == NULL )
	{
		idx++;
	}
	
	if ( idx < DICT_LENGTH )
	{
		*handle = idx;
		NutMutexUnlock( &dict_mutex );			
		return &packet_dict[idx];
	}	
	
	NutMutexUnlock( &dict_mutex );		
	return NULL;
}

// ----------------------------------------------------------------------------
// ADC Stuff
// ----------------------------------------------------------------------------

u_short adc2_init( adc2_mode_t mode, 
				adc2_interrupt_t use_interrupt, 
				adc2_prescaler_t prescaler, 
				adc2_channel_t channel, 
				adc2_ref_t reference )
{
	u_char admux_new = 0;
	u_char adcsr_new = 0;
	
    adcsr_new |= mode;
    adcsr_new |= use_interrupt;
    adcsr_new |= prescaler;
    
    admux_new |= channel;
    admux_new |= reference;   		
	
	NutEnterCritical();
	
	/* disable ADC */
    cbi(ADCSR, ADEN);
  
//    /* perform only one single conversion at a time */
//    cbi(ADCSR, ADFR);
//    /* disable ADC interrupt */
//    cbi(ADCSR, ADIE);
//    /* set ADC prescaler to DIV2 */
//    cbi(ADCSR, ADPS2);
//    cbi(ADCSR, ADPS1);
//    cbi(ADCSR, ADPS0);
//    /* set ADC channel */
//    ADMUX = (ADMUX & 0xf8) | 0;
//    /* use external voltage reference */
//    cbi(ADMUX, REFS1);
//    cbi(ADMUX, REFS0);
    
    ADCSR = adcsr_new;
    ADMUX = admux_new;
    
    /* enable ADC */
    sbi(ADCSR, ADEN);
    
   	NutExitCritical();
   	
   	//printf( "adc2_init(); admux_new = %02X, adcsr_new = %02X\n", admux_new, adcsr_new );
   	
   	return admux_new << 8 | adcsr_new;
}

u_short adc2_read( u_short handle )
{
	u_char  admux_new = handle >> 8;
	u_char  adcsr_new = handle & 0xFF; // | 1 << 7;
	u_char  admux_old = 0;
	u_char  adcsr_old = 0;
	u_short result = 0;
	
	//printf( "Handle = %04x\n", handle );
	
	NutEnterCritical();
	
	admux_old = ADMUX; // mask out Bits 7 and 4
	adcsr_old = ADCSR & 0x6F;
	
	// check if the ADC registers have changed...
	if ( admux_old != admux_new || adcsr_old != adcsr_new )
	{
		//printf( "Updating ADC... %02X!=%02X,%02X!=%02X\n", 
		//admux_old, admux_new, adcsr_old, adcsr_new );
		
		/* disable ADC */
	    cbi(ADCSR, ADEN);		
	    
	    ADCSR = adcsr_new;
	    ADMUX = admux_new;
	    
	    /* enable ADC */
	    sbi(ADCSR, ADEN);
	}
	
	sbi(ADCSR, ADSC);
    while (bit_is_set(ADCSR, ADSC));
	result = ADCL | (ADCH << 8);  	

   	NutExitCritical();
   	
   	return result;
}

